# -*- coding: utf-8 -*-
import sys

import canmatrix.formats
import pytest

try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path

pytest_plugins = ["pytester"]
here = Path(__file__).parent


@pytest.fixture
def run(testdir):
    def do_run(*args):
        args = [sys.executable,"-m","canmatrix.cli.convert"] + list(args)
        return testdir.run(*args)
    return do_run

def test_silent(tmpdir, run):
    inputFile = str(here / "test_frame_decoding.dbc")

    normal_result = run(inputFile ,"tmp.dbc")
    silent_result = run("-s", inputFile,"tmp.dbc")
    assert len(normal_result.errlines) > len(silent_result.errlines)

def test_verbose(tmpdir, run):
    inputFile = str(here / "ARXML_min_max.arxml")

    normal_result = run(inputFile ,"tmp.dbc")
    verbose_result = run("-vv", inputFile,"tmp.dbc")
    assert len(normal_result.errlines) < len(verbose_result.errlines)

def test_force_output_format(tmpdir, run):
    inputFile = str(here / "test_frame_decoding.dbc")
    outFile =  str(here / "tmp.tmp")
    normal_result = run("-v", "-f","dbc", inputFile, outFile)
    assert 'INFO - convert - done' in normal_result.errlines[-1]
    with open(outFile, "r") as fd:
        first_line = fd.readline()
        assert first_line == 'VERSION "created by canmatrix"\n'

def test_foce_input_format(tmpdir, run):
    #requires test_force_output to run first
    inputFile =  str(here / "tmp.tmp")
    normal_result = run("-i","dbc", inputFile, "tmp.dbc")
    assert 'INFO - convert - done' in normal_result.errlines[-1]

def create_dbc_with_special_char():
    outFile = str(here / "tmp.dbc")
    myFrame = canmatrix.Frame("testFrame1", arbitration_id=canmatrix.arbitration_id_converter(0x123), size=8, transmitters=["testBU"])
    mySignal = canmatrix.Signal("someTestSignal",
                      size=11,
                      is_little_endian=False,
                      is_signed=False,
                      factor=5.0,
                      offset=1.0,
                      min=0,
                      max=500,
                      unit=u"specialCharUnit°$",  # .decode("utf-8"),
                      receivers=["recBU"])
    myFrame.add_signal(mySignal)

    db = canmatrix.CanMatrix()
    db.add_frame(myFrame)
    db.add_frame_defines("SomeUnneededDefine", 'INT 0 65535')
    canmatrix.formats.dumpp({"": db}, outFile, dbcExportEncoding='iso-8859-1',
                            dbcExportCommentEncoding='iso-8859-1')
    return outFile

def test_ignore_encoding_errors(tmpdir, run):
    inputFile = create_dbc_with_special_char()
    normal_result = run("--ignoreEncodingErrors","--dbcExportEncoding", "ascii", inputFile, "tmp2.dbc")
    assert 'INFO - convert - done' in normal_result.errlines[-1]

def test_delete_obsolete_defines(tmpdir, run):
    inputFile = create_dbc_with_special_char()
    deleted_result = run("--deleteObsoleteDefines", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","rb") as fd:
        content = fd.read()
        assert b"BA_DEF_" not in content
    normal_result = run(inputFile, "tmp2.dbc")
    with open("tmp2.dbc","rb") as fd:
        content = fd.read()
        assert b"BA_DEF_" in content

def test_delete_ecu(tmpdir, run):
    inputFile = create_dbc_with_special_char()
    deleted_result = run("--deleteEcu","testBU", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","rb") as fd:
        content = fd.read()
        assert b"testBU" not in content

def test_rename_ecu(tmpdir, run):
    inputFile = create_dbc_with_special_char()
    deleted_result = run("--renameEcu","testBU:renamedECU", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","rb") as fd:
        content = fd.read()
        assert b"testBU" not in content
        assert b"renamedECU" in content

def test_delete_signal(tmpdir, run):
    inputFile = create_dbc_with_special_char()
    deleted_result = run("--deleteSignal","someTestSignal", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","rb") as fd:
        content = fd.read()
        assert b"someTestSignal" not in content

def test_rename_signal(tmpdir, run):
    inputFile = create_dbc_with_special_char()
    deleted_result = run("--renameSignal","someTestSignal:renamedSignal", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","rb") as fd:
        content = fd.read()
        assert b"someTestSignal" not in content
        assert b"renamedSignal" in content

def test_delete_frame(tmpdir, run):
    inputFile = create_dbc_with_special_char()
    deleted_result = run("--deleteFrame","testFrame1", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","r") as fd:
        content = fd.read()
        assert "testFrame1" not in content

def test_rename_frame(tmpdir, run):
    inputFile = create_dbc_with_special_char()
    deleted_result = run("--renameFrame","testFrame1:renamedFrame", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","rb") as fd:
        content = fd.read()
        assert b"testFrame1" not in content
        assert b"renamedFrame" in content

def test_add_frame_receiver(tmpdir, run):
    inputFile = create_dbc_with_special_char()
    deleted_result = run("--addFrameReceiver","testFrame1:newECU", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","rb") as fd:
        content = fd.read()
        assert b"recBU,newECU" in content

def test_change_frame_id(tmpdir, run):
    inputFile = create_dbc_with_special_char()
    deleted_result = run("--changeFrameId","291:666", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","rb") as fd:
        content = fd.read()
        assert b"BO_ 666" in content

def test_set_frame_fd(tmpdir, run):
    inputFile = create_dbc_with_special_char()
    deleted_result = run("--setFrameFd","testFrame1", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","rb") as fd:
        content = fd.read()
        assert b'BA_ "VFrameFormat" BO_ 291 14' in content
    deleted_result = run("--unsetFrameFd","testFrame1", "tmp2.dbc", "tmp3.dbc")
    with open("tmp3.dbc","rb") as fd:
        content = fd.read()
        assert b'BA_ "VFrameFormat" BO_ 291 14' not in content

def test_recalc_dlc(tmpdir, run):
    inputFile = create_dbc_with_special_char()
    result = run("--recalcDLC","max", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","rb") as fd:
        content = fd.read()
        assert b"testFrame1: 8" in content

    result = run("--recalcDLC","force", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","rb") as fd:
        content = fd.read()
        assert b"testFrame1: 2" in content

def test_skip_long_dlc(tmpdir, run):
    inputFile = create_dbc_with_special_char()
    result = run("--skipLongDlc", "7", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","r") as fd:
        content = fd.read()
        assert "someTestSignal" not in content

def test_cut_long_frames(tmpdir, run):
    inputFile = create_dbc_with_special_char()
    result = run("--cutLongFrames", "1", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","r") as fd:
        content = fd.read()
        assert "someTestSignal" not in content
    result = run("--cutLongFrames", "2", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","rb") as fd:
        content = fd.read()
        assert b"someTestSignal" in content

def test_copy_signals(tmpdir, run):
    inputFile = create_dbc_with_special_char()
    result = run("--signals", "someTestSignal", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","rb") as fd:
        content = fd.read()
        assert b"someTestSignal" in content
        assert b"VECTOR__INDEPENDENT_SIG_MSG" in content


def create_dbc(additionalReceiver = []):
    outFile = str(here / "tmpb.dbc")
    myFrame = canmatrix.Frame("testFrame3", arbitration_id=canmatrix.arbitration_id_converter(0x124), size=8, transmitters=["testBU"])
    mySignal = canmatrix.Signal("someTestSignal",
                      size=11,
                      is_little_endian=False,
                      is_signed=False,
                      factor=5.0,
                      offset=1.0,
                      min=0,
                      max=500,
                      receivers=["recBU"])
    myFrame.add_signal(mySignal)
    myFrame2 = canmatrix.Frame("testFrame2", arbitration_id=canmatrix.arbitration_id_converter(0x125), size=8, transmitters=["testBU2"])
    myFrame2.add_attribute("myAttribute","42")
    mySignal2 = canmatrix.Signal("someTestSignal2",
                      start_bit=15,
                      size=11,
                      is_little_endian=False,
                      is_signed=False,
                      factor=5.0,
                      offset=1.0,
                      min=0,
                      max=500,
                      receivers=["recBU2"] + additionalReceiver)
    myFrame2.add_signal(mySignal2)
    mySignal3 = canmatrix.Signal("zeroSignal",
                      start_bit=20,
                      size=0,
                      is_little_endian=False,
                      is_signed=False,
                      factor=5.0,
                      offset=1.0,
                      min=0,
                      max=500,
                      receivers=["recBU2"])
    mySignal3.add_attribute("mySignalAttribute", "7")
    myFrame2.add_signal(mySignal3)

    db = canmatrix.CanMatrix()
    db.add_frame(myFrame)
    db.add_frame(myFrame2)
    db.add_frame_defines("myAttribute","INT -5 10")
    db.add_signal_defines("mySignalAttribute", 'INT 0 65535')
    canmatrix.formats.dumpp({"": db}, outFile, dbcExportEncoding='iso-8859-1',
                            dbcExportCommentEncoding='iso-8859-1')
    return outFile

def test_copy_ecus(tmpdir, run):
    inputFile = create_dbc()
    result = run("--ecus", "testBU", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","r") as fd:
        content = fd.read()
        assert "testBU2" not in content
        assert "testBU" in content

def test_copy_ecus_rx(tmpdir, run):
    inputFile = create_dbc()
    result = run("--ecus", "recBU:rx", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","r") as fd:
        content = fd.read()
        assert "recBU2" not in content
        assert "recBU" in content

def test_copy_ecus_tx(tmpdir, run):
    inputFile = create_dbc(additionalReceiver = ["testBU"])
    result = run("--ecus", "testBU:tx", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","r") as fd:
        content = fd.read()
        assert "testFrame2" not in content
        assert "testFrame3" in content

def test_copy_frames(tmpdir, run):
    inputFile = create_dbc()
    result = run("--frames", "testFrame3", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","rb") as fd:
        content = fd.read()
        assert b"testFrame2" not in content
        assert b"testFrame3" in content

def test_delete_frame_attributes(tmpdir, run):
    inputFile = create_dbc()
    result = run("--deleteFrameAttributes", "myAttribute", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","r") as fd:
        content = fd.read()
        assert 'BA_ "myAttribute"' not in content

def test_delete_zero_signals(tmpdir, run):
    inputFile = create_dbc()
    result = run("--deleteZeroSignals", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","r") as fd:
        content = fd.read()
        assert 'zeroSignal' not in content


def test_delete_signal_attributes(tmpdir, run):
    inputFile = create_dbc()
    result = run("--deleteSignalAttributes", "mySignalAttribute", inputFile, "tmp2.dbc")
    with open("tmp2.dbc","r") as fd:
        content = fd.read()
        assert 'BA_ "mySignalAttribute"' not in content

def test_merge(tmpdir, run):
    inputFile1 = create_dbc_with_special_char()
    inputFile2 = create_dbc()

    if inputFile1[1] == ":":
        inputFile1 = inputFile1[2:]
    result = run("--merge", inputFile1, inputFile2, "tmp2.dbc")
    with open("tmp2.dbc", "rb") as fd:
        content = fd.read()
        assert b"BO_ 291" in content
        assert b"BO_ 292" in content
        assert b"BO_ 293" in content

    result = run("--merge", inputFile1 + ":ecu=testBU", inputFile2, "tmp3.dbc")
    with open("tmp3.dbc", "rb") as fd:
        content = fd.read()
        assert b"BO_ 291" in content
        assert b"BO_ 292" in content
        assert b"BO_ 293" in content

    result = run("--merge", inputFile1 + ":frame=testFrame1", inputFile2, "tmp4.dbc")
    with open("tmp4.dbc", "rb") as fd:
        content = fd.read()
        assert b"BO_ 291" in content
        assert b"BO_ 292" in content
        assert b"BO_ 293" in content



# -*- coding: utf-8 -*-
# Copyright (c) 2013, Eduard Broecker
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that
# the following conditions are met:
#
#    Redistributions of source code must retain the above copyright notice, this list of conditions and the
#    following disclaimer.
#    Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

from __future__ import absolute_import, division, print_function

import logging
import sys
import typing
from builtins import *

import attr

import canmatrix

logger = logging.getLogger(__name__)
ConfigDict = typing.Optional[typing.Mapping[str, typing.Union[str, bool]]]
WithAttribute = typing.TypeVar("WithAttribute", canmatrix.CanMatrix, canmatrix.Ecu, canmatrix.Frame, canmatrix.Signal)


@attr.s
class CompareResult(object):
    """Hold comparison results in logical tree."""
    result = attr.ib(default=None)  # type: typing.Optional[str]    # any of equal, added, deleted, changed
    type = attr.ib(default=None)  # type: typing.Optional[str]      # db, ecu, frame, signal, signalGroup or attribute
    ref = attr.ib(default=None)  # type: typing.Any                 # reference to related object
    changes = attr.ib(default=None)  # type: typing.Optional[typing.List]
    _children = attr.ib(factory=list)  # type: typing.List[CompareResult]  # nested CompareResults

    def add_child(self, child):
        # type: (CompareResult) -> None
        self._children.append(child)

    @property
    def children(self):  # type: () -> typing.List[CompareResult]
        return self._children


def propagate_changes(res):  # type: (CompareResult) -> int
    change = 0
    for child in res.children:
        change += propagate_changes(child)
    if change != 0:
        res.result = "changed"
    if res.result != "equal":
        return 1
    else:
        return 0


def compare_db(db1, db2, ignore=None):
    # type: (canmatrix.CanMatrix, canmatrix.CanMatrix, ConfigDict) -> CompareResult
    result = CompareResult()
    if ignore is None:
        ignore = dict()
    for f1 in db1.frames:
        f2 = db2.frame_by_name(f1.name)
        f2id = db2.frame_by_id(f1.arbitration_id)
        if f2 is None:
            if f2id is None:
                result.add_child(CompareResult("deleted", "FRAME", f1))
            else:
                result.add_child(compare_frame(f1, f2id, ignore))
        else:
            result.add_child(compare_frame(f1, f2, ignore))
    for f2 in db2.frames:
        f1 = db1.frame_by_name(f2.name)
        f1id = db1.frame_by_id(f2.arbitration_id)
        if f1id is None and f1 is None:
            result.add_child(CompareResult("added", "FRAME", f2))

    if "ATTRIBUTE" in ignore and ignore["ATTRIBUTE"] == "*":
        pass
    else:
        result.add_child(compare_attributes(db1, db2, ignore))

    for ecu1 in db1.ecus:
        ecu2 = db2.ecu_by_name(ecu1.name)
        if ecu2 is None:
            result.add_child(CompareResult("deleted", "ecu", ecu1))
        else:
            result.add_child(compare_ecu(ecu1, ecu2, ignore))
    for ecu2 in db2.ecus:
        ecu1 = db1.ecu_by_name(ecu2.name)
        if ecu1 is None:
            result.add_child(CompareResult("added", "ecu", ecu2))

    if "DEFINE" in ignore and ignore["DEFINE"] == "*":
        pass
    else:
        result.add_child(
            compare_define_list(
                db1.global_defines,
                db2.global_defines))

        temp = compare_define_list(db1.ecu_defines, db2.ecu_defines)
        temp.type = "ECU Defines"
        result.add_child(temp)

        temp = compare_define_list(db1.frame_defines, db2.frame_defines)
        temp.type = "Frame Defines"
        result.add_child(temp)

        temp = compare_define_list(db1.signal_defines, db2.signal_defines)
        temp.type = "Signal Defines"
        result.add_child(temp)

    if "VALUETABLES" in ignore and ignore["VALUETABLES"]:
        pass
    else:
        for vt1 in db1.value_tables:
            if vt1 not in db2.value_tables:
                result.add_child(
                    CompareResult(
                        "deleted",
                        "valuetable " + vt1,
                        db1.value_tables))
            else:
                result.add_child(
                    compare_value_table(
                        db1.value_tables[vt1],
                        db2.value_tables[vt1]))

        for vt2 in db2.value_tables:
            if vt2 not in db1.value_tables:
                result.add_child(
                    CompareResult(
                        "added",
                        "valuetable " + vt2,
                        db2.value_tables))

    propagate_changes(result)

    return result


def compare_value_table(vt1, vt2):
    # type: (typing.Mapping, typing.Mapping) -> CompareResult
    result = CompareResult("equal", "Valuetable", vt1)
    for value in vt1:
        if value not in vt2:
            result.add_child(
                CompareResult(
                    "removed",
                    "Value " +
                    str(value),
                    vt1[value]))
        elif vt1[value] != vt2[value]:
            result.add_child(CompareResult("changed", "Value " +
                                           str(value) +
                                           " " +
                                           str(vt1[value].encode('ascii', 'ignore')), [vt1[value], vt2[value]]))
    for value in vt2:
        if value not in vt1:
            result.add_child(
                CompareResult(
                    "added",
                    "Value " +
                    str(value),
                    vt2[value]))
    return result


def compare_signal_group(sg1, sg2):
    # type: (canmatrix.SignalGroup, canmatrix.SignalGroup) -> CompareResult
    result = CompareResult("equal", "SignalGroup", sg1)

    if sg1.name != sg2.name:
        result.add_child(
            CompareResult(
                "changed", "SignalName", [
                    sg1.name, sg2.name]))
    if sg1.id != sg2.id:
        result.add_child(CompareResult(
            "changed", "SignalName", [str(sg1.id), str(sg2.id)]))

    if sg1.signals is None or sg2.signals is None:
        logger.debug("Strange - sg wo members???")
        return result
    for signal in sg1.signals:
        if sg2.by_name(signal.name) is None:
            result.add_child(CompareResult("deleted", str(signal.name), signal))
    for signal in sg2.signals:
        if sg1.by_name(signal.name) is None:
            result.add_child(CompareResult("added", str(signal.name), signal))
    return result


def compare_define_list(d1list, d2list):
    # type: (typing.Mapping[str, canmatrix.Define], typing.Mapping[str, canmatrix.Define]) -> CompareResult
    result = CompareResult("equal", "DefineList", d1list)
    for definition in d1list:
        if definition not in d2list:
            result.add_child(
                CompareResult(
                    "deleted",
                    "Define" +
                    str(definition),
                    d1list))
        else:
            d2 = d2list[definition]
            d1 = d1list[definition]
            if d1.definition != d2.definition:
                result.add_child(
                    CompareResult(
                        "changed", "Definition", d1.definition, [
                            d1.definition, d2.definition]))

            if d1.defaultValue != d2.defaultValue:
                result.add_child(
                    CompareResult(
                        "changed", "DefaultValue", d1.definition, [
                            d1.defaultValue, d2.defaultValue]))
    for definition in d2list:
        if definition not in d1list:
            result.add_child(
                CompareResult(
                    "added",
                    "Define" +
                    str(definition),
                    d2list))
    return result


def compare_attributes(ele1, ele2, ignore=None):
    # type: (WithAttribute, WithAttribute, ConfigDict) -> CompareResult
    if ignore is None:
        ignore = dict()
    result = CompareResult("equal", "ATTRIBUTES", ele1)
    if "ATTRIBUTE" in ignore and (
            ignore["ATTRIBUTE"] == "*" or ignore["ATTRIBUTE"] == ele1):
        return result
    for attribute in ele1.attributes:
        if attribute not in ele2.attributes:
            result.add_child(
                CompareResult(
                    "deleted",
                    str(attribute),
                    ele1.attributes[attribute]))
        elif ele1.attributes[attribute] != ele2.attributes[attribute]:
            result.add_child(
                CompareResult(
                    "changed", str(attribute), ele1.attributes[attribute], [
                        ele1.attributes[attribute], ele2.attributes[attribute]]))

    for attribute in ele2.attributes:
        if attribute not in ele1.attributes:
            result.add_child(
                CompareResult(
                    "added",
                    str(attribute),
                    ele2.attributes[attribute]))
    return result


def compare_ecu(ecu1, ecu2, ignore=None):
    # type: (canmatrix.Ecu, canmatrix.Ecu, ConfigDict) -> CompareResult
    if ignore is None:
        ignore = dict()
    result = CompareResult("equal", "ECU", ecu1)

    if "comment" not in ignore:
        if ecu1.comment != ecu2.comment:
            result.add_child(
                CompareResult(
                    "changed", "ECU", ecu1, [
                        ecu1.comment, ecu2.comment]))

    if "ATTRIBUTE" in ignore and ignore["ATTRIBUTE"] == "*":
        pass
    else:
        result.add_child(compare_attributes(ecu1, ecu2, ignore))
    return result


def compare_frame(f1, f2, ignore=None):
    # type: (canmatrix.Frame, canmatrix.Frame, ConfigDict) -> CompareResult
    if ignore is None:
        ignore = dict()
    result = CompareResult("equal", "FRAME", f1)

    for s1 in f1:
        s2 = f2.signal_by_name(s1.name)
        if not s2:
            result.add_child(CompareResult("deleted", "SIGNAL", s1))
        else:
            result.add_child(compare_signal(s1, s2, ignore))

    if f1.name != f2.name:
        result.add_child(
            CompareResult(
                "changed", "Name", f1, [
                    f1.name, f2.name]))
    if f1.size != f2.size:
        result.add_child(
            CompareResult(
                "changed", "dlc", f1, [
                    "dlc: %d" %
                    f1.size, "dlc: %d" %
                    f2.size]))
    if f1.arbitration_id.id != f2.arbitration_id.id:
        result.add_child(
            CompareResult(
                "changed", "ID", f1, [
                    "ID: %xh" %
                    f1.arbitration_id.id, "ID: %xh" %
                    f2.arbitration_id.id]))
    if f1.arbitration_id.extended != f2.arbitration_id.extended:
        result.add_child(
            CompareResult(
                "changed", "FRAME", f1, [
                    "extended-Flag: %d" %
                    f1.arbitration_id.extended, "extended-Flag: %d" %
                    f2.arbitration_id.extended]))
    if "comment" not in ignore:
        if f2.comment is None:
            f2.add_comment("")
        if f1.comment is None:
            f1.add_comment("")
        if f1.comment != f2.comment:
            result.add_child(
                CompareResult(
                    "changed", "FRAME", f1, [
                        "comment: " + f1.comment, "comment: " + f2.comment]))

    for s2 in f2.signals:
        s1 = f1.signal_by_name(s2.name)
        if not s1:
            result.add_child(CompareResult("added", "SIGNAL", s2))

    if "ATTRIBUTE" in ignore and ignore["ATTRIBUTE"] == "*":
        pass
    else:
        result.add_child(compare_attributes(f1, f2, ignore))

    temp = [str(item) for item in f2.transmitters]
    for transmitter in f1.transmitters:
        if transmitter not in temp:
            result.add_child(CompareResult("removed", "Frame-Transmitter", f1))

    temp = [str(item) for item in f1.transmitters]
    for transmitter in f2.transmitters:
        if transmitter not in temp:
            result.add_child(CompareResult("added", "Frame-Transmitter", f2))

    for sg1 in f1.signalGroups:
        sg2 = f2.signal_group_by_name(sg1.name)
        if sg2 is None:
            result.add_child(CompareResult("removed", "Signalgroup", sg1))
        else:
            result.add_child(compare_signal_group(sg1, sg2))

    for sg2 in f2.signalGroups:
        if f1.signal_group_by_name(sg2.name) is None:
            result.add_child(CompareResult("added", "Signalgroup", sg2))
    return result


def compare_signal(s1, s2, ignore=None):
    # type: (canmatrix.Signal, canmatrix.Signal, ConfigDict) -> CompareResult
    if ignore is None:
        ignore = dict()
    result = CompareResult("equal", "SIGNAL", s1)

    if s1.start_bit != s2.start_bit:
        result.add_child(
            CompareResult(
                "changed", "startbit", s1, [
                    " %d" %
                    s1.start_bit, " %d" %
                    s2.start_bit]))
    if s1.size != s2.size:
        result.add_child(
            CompareResult(
                "changed", "signalsize", s1, [
                    " %d" %
                    s1.size, " %d" %
                    s2.size]))
    if float(s1.factor) != float(s2.factor):
        result.add_child(
            CompareResult(
                "changed", "factor", s1, [
                    s1.factor, s2.factor]))
    if float(s1.offset) != float(s2.offset):
        result.add_child(
            CompareResult(
                "changed", "offset", s1, [
                    s1.offset, s2.offset]))
    if float(s1.min) != float(s2.min):
        result.add_child(
            CompareResult(
                "changed", "min", s1, [
                    s1.min, s2.min]))
    if float(s1.max) != float(s2.max):
        result.add_child(
            CompareResult(
                "changed", "max", s1, [
                    s1.max, s2.max]))
    if s1.is_little_endian != s2.is_little_endian:
        result.add_child(
            CompareResult(
                "changed", "is_little_endian", s1, [
                    " %d" %
                    s1.is_little_endian, " %d" %
                    s2.is_little_endian]))
    if s1.is_signed != s2.is_signed:
        result.add_child(
            CompareResult(
                "changed", "sign", s1, [
                    " %d" %
                    s1.is_signed, " %d" %
                    s2.is_signed]))
    if s1.multiplex != s2.multiplex:
        result.add_child(CompareResult("changed", "multiplex", s1, [
            str(s1.multiplex), str(s2.multiplex)]))
    if s1.unit != s2.unit:
        result.add_child(
            CompareResult(
                "changed", "unit", s1, [
                    s1.unit, s2.unit]))
    if "comment" not in ignore:
        if s1.comment is not None and s2.comment is not None and s1.comment != s2.comment:
            if s1.comment.replace("\n", " ") != s2.comment.replace("\n", " "):
                result.add_child(
                    CompareResult(
                        "changed", "comment", s1, [
                            s1.comment, s2.comment]))
            else:
                result.add_child(
                    CompareResult(
                        "changed", "comment", s1, [
                            "only whitespaces differ", ""]))

    for receiver in s1.receivers:
        if receiver.strip() not in s2.receivers:
            result.add_child(
                CompareResult(
                    "removed",
                    "receiver " +
                    receiver,
                    s1.receivers))

    for receiver in s2.receivers:
        if receiver.strip() not in s1.receivers:
            result.add_child(
                CompareResult(
                    "added",
                    "receiver " +
                    receiver,
                    s1.receivers))

    if "ATTRIBUTE" in ignore and ignore["ATTRIBUTE"] == "*":
        pass
    else:
        result.add_child(compare_attributes(s1, s2, ignore))

    if "VALUETABLES" in ignore and ignore["VALUETABLES"]:
        pass
    else:
        result.add_child(compare_value_table(s1.values, s2.values))

    return result


def dump_result(res, depth=0):
    # type: (CompareResult, int) -> None
    if res.type is not None and res.result != "equal":
        for _ in range(0, depth):
            print(" ", end=' ')
        print(res.type + " " + res.result + " ", end=' ')
        if hasattr(res.ref, 'name'):
            print(res.ref.name)
        else:
            print(" ")
        if res.changes is not None and res.changes[0] is not None and res.changes[1] is not None:
            for _ in range(0, depth):
                print(" ", end=' ')
            print(type(res.changes[0]))
            if sys.version_info[0] < 3:
                if isinstance(res.changes[0], type(u'')):
                    res.changes[0] = res.changes[0].encode('ascii', 'ignore')
                if isinstance(res.changes[1], type(u'')):
                    res.changes[1] = res.changes[1].encode('ascii', 'ignore')
            else:
                if type(res.changes[0]) == str:
                    res.changes[0] = res.changes[0].encode('ascii', 'ignore')
                if type(res.changes[1]) == str:
                    res.changes[1] = res.changes[1].encode('ascii', 'ignore')
            print("old: " +
                  str(res.changes[0]) +
                  " new: " +
                  str(res.changes[1]))
    for child in res.children:
        dump_result(child, depth + 1)


# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from builtins import *

import attr

import canmatrix.formats

try:
    from importlib.resources import read_binary
except ImportError:
    from pkgutil import get_data as read_binary


@attr.s
class j1939_decoder(object):
    string = read_binary(__name__.rpartition('.')[0], "j1939.dbc")
    j1939_db = canmatrix.formats.loads_flat(
        string, import_type="dbc", dbcImportEncoding="utf8"
    )
    length = attr.ib(default=0)  # type: int
    count_succesive_frames = attr.ib(default=0)  # type: int
    transfered_pgn = attr.ib(default=0)  # type: int
    _data = attr.ib(init=False, default=bytearray())

    def decode(self, arbitration_id, can_data, matrix = None):
        if matrix is not None:
            frame = matrix.frame_by_pgn(arbitration_id.pgn)
        else:
            frame = None
        if frame is not None:
            return ("regular " + frame.name, frame.decode(can_data))
        elif self.j1939_db.frame_by_pgn(arbitration_id.pgn) is not None:
            signals = self.j1939_db.decode(arbitration_id,can_data)
            frame_name = self.j1939_db.frame_by_pgn(arbitration_id.pgn).name
            return ("J1939 known: " + frame_name, signals)

        elif arbitration_id.pgn == canmatrix.ArbitrationId.from_pgn(0xECFF).pgn and can_data[0] == 32:
            # BAM detected
            self.length = (int(can_data[2]) << 8) + int(can_data[1])
            self.count_succesive_frames = int(can_data[3])
            self.transfered_pgn = (int(can_data[7]) << 16) + (int(can_data[6]) << 8) + int(can_data[5])
            self.bytes_left = self.length
            self._data = bytearray()
            return ("BAM          ", {})

        elif arbitration_id.pgn == canmatrix.ArbitrationId.from_pgn(0xECFF).pgn and can_data[0] == 16:
            # RTS detected
            self.length = (int(can_data[2]) << 8) + int(can_data[1])
            self.count_of_packets = int(can_data[3])
            self.total_count_of_packet_sent = int(can_data[4])
            self.transfered_pgn = (int(can_data[7]) << 16) + (int(can_data[6]) << 8) + int(can_data[5])
            return ("ERROR - decoding RTS not yet implemented")

        elif arbitration_id.pgn == canmatrix.ArbitrationId.from_pgn(0xECFF).pgn and can_data[0] == 17:
            # CTS detected
            self.max_packets_at_once = can_data[1]
            self.sequence_number_to_start = can_data[2]
            self.transfered_pgn = (int(can_data[7]) << 16) + (int(can_data[6]) << 8) + int(can_data[5])
            return ("ERROR - decoding CTS not yet implemented")

        elif arbitration_id.pgn == canmatrix.ArbitrationId.from_pgn(0xECFF).pgn and can_data[0] == 19:
            # ACK detected
            self.message_size = (int(can_data[2]) << 8) + int(can_data[1])
            self.count_of_packets = int(can_data[3])
            self.transfered_pgn = (int(can_data[7]) << 16) + (int(can_data[6]) << 8) + int(can_data[5])
            return ("ERROR - decoding ACK not yet implemented")

        elif arbitration_id.pgn == canmatrix.ArbitrationId.from_pgn(0xECFF).pgn and can_data[0] == 255:
            # Connection Abort
            self.abort_reason = can_data[1]
            self.transfered_pgn = (int(can_data[7]) << 16) + (int(can_data[6]) << 8) + int(can_data[5])
            return ("ERROR - decoding Connection Abbort not yet implemented")


        elif arbitration_id.pgn == canmatrix.ArbitrationId.from_pgn(0xEEFF).pgn:
            #Address Claimed
            #arbitration_id.j1939_source
            #name in can_data[0:8]
            return ("ERROR - address claim detected not yet implemented")
            pass

        elif arbitration_id.pgn == canmatrix.ArbitrationId.from_pgn(0xEBFF).pgn:
            # transfer data
            self._data = self._data + can_data[1:min(8, self.bytes_left + 1)]
            self.bytes_left = max(self.bytes_left - 7, 0)

            if self.count_succesive_frames == 0:
                #print(self._data)
                frame = matrix.frame_by_pgn(self.transfered_pgn)
                if frame is not None:
                    signals = frame.decode(self._data)
                    return ("BAM last data", signals)
                return ("BAM last data", {})
            else:
                self.count_succesive_frames -= 1
                return ("BAM data     ", {})
        return ("",{})



